import com.atlassian.jira.bc.JiraServiceContextImpl
import com.atlassian.jira.component.ComponentAccessor
import com.atlassian.jira.issue.fields.AbstractProgressBarSystemField
import com.atlassian.jira.issue.fields.layout.column.ColumnLayoutItemImpl
import com.atlassian.jira.issue.views.util.IssueViewUtil
import com.atlassian.jira.issue.worklog.WorkRatio
import com.atlassian.jira.timezone.TimeZoneService
import com.midori.jira.plugin.commons.util.ClassHelper
import com.midori.jira.plugin.commons.util.VersionHelper
import org.apache.log4j.Logger

fieldHelper = new FieldHelperTool(user, requestContext.canonicalBaseUrl, i18n)

/**
 * Tool to convert Jira system- and custom fields to their best Excel representations.
 */
public class FieldHelperTool {
	def log = Logger.getLogger(this.getClass())

	/* Configuration. */
	def issueKeyAsLink = true
	def summaryAsLink = true

	/* Jira Service Desk configuration. */
	def sdExportCustomerSatisfactionComment = false // requires Service Desk 3.6.0+

	/* Insight configuration. */
	def insightObjectsWithNames = true
	def insightObjectsWithKeys = true
	def insightObjectsWithAttributes = false

	/* Xray configuration. */
	def xrayIssuesWithSummaries = true
	def xray4OrLater = !ComponentAccessor.customFieldManager.getCustomFieldType('com.xpandit.plugins.xray:manual-test-steps-export-custom-field') // "Manual Test Steps Export" custom field was removed in Xray 4.0

	/* Dependencies. */
	def commentManager = ComponentAccessor.commentManager
	def customFieldManager = ComponentAccessor.customFieldManager
	def issueManager = ComponentAccessor.issueManager
	def issueViewUtil = ComponentAccessor.getOSGiComponentInstanceOfType(IssueViewUtil.class)
	def timeZoneService = ComponentAccessor.getOSGiComponentInstanceOfType(TimeZoneService.class)

	def workRatio = new WorkRatio()

	/* Jira Service Desk dependencies. */
	def issueIdToGoalDetailedViewDataList = [:]
	def requestFeedbackInternalServiceInitialized = false
	def requestFeedbackInternalService

	/* Insight dependencies. */
	def insightObjectFacade
	def insightObjectTypeAttributeFacade

	/* nFeed dependencies. */
	def nFeedInitialized = false
	def nFeedFieldDisplayService

	/* ScriptRunner dependencies. */
	def scriptFieldConfigurations
	def scriptFieldIdsWithCustomTemplate
	def fieldIdToSearcherKey = [:]

	def user
	def baseUrl
	def calendar

	def i18n

	private static EXCEL_PARENT_KEY_FIELD_NAME = "Excel parent key"
	private static EXCEL_COMMENTS_CALCULATED_FIELD_NAME = "Excel comments"
	private static EXCEL_FIRST_COMMENT_CALCULATED_FIELD_NAME = "Excel first comment"
	private static EXCEL_FIRST_COMMENT_TEXT_CALCULATED_FIELD_NAME = "Excel first comment text"
	private static EXCEL_FIRST_COMMENT_AUTHOR_CALCULATED_FIELD_NAME = "Excel first comment user"
	private static EXCEL_FIRST_COMMENT_DATE_CALCULATED_FIELD_NAME = "Excel first comment date"
	private static EXCEL_LAST_COMMENT_CALCULATED_FIELD_NAME = "Excel last comment"
	private static EXCEL_LAST_COMMENT_TEXT_CALCULATED_FIELD_NAME = "Excel last comment text"
	private static EXCEL_LAST_COMMENT_AUTHOR_CALCULATED_FIELD_NAME = "Excel last comment user"
	private static EXCEL_LAST_COMMENT_DATE_CALCULATED_FIELD_NAME = "Excel last comment date"

	private static CALCULATED_FIELD_NAMES = [
			EXCEL_PARENT_KEY_FIELD_NAME,
			EXCEL_COMMENTS_CALCULATED_FIELD_NAME,
			EXCEL_FIRST_COMMENT_CALCULATED_FIELD_NAME, EXCEL_FIRST_COMMENT_TEXT_CALCULATED_FIELD_NAME, EXCEL_FIRST_COMMENT_AUTHOR_CALCULATED_FIELD_NAME, EXCEL_FIRST_COMMENT_DATE_CALCULATED_FIELD_NAME,
			EXCEL_LAST_COMMENT_CALCULATED_FIELD_NAME, EXCEL_LAST_COMMENT_TEXT_CALCULATED_FIELD_NAME, EXCEL_LAST_COMMENT_AUTHOR_CALCULATED_FIELD_NAME, EXCEL_LAST_COMMENT_DATE_CALCULATED_FIELD_NAME
	]

	/**
	 * Cell values using structured representation are indented with this.
	 */
	private static INDENTATION = "    "

	public FieldHelperTool(def user, def baseUrl, def i18n) {
		this.user = user
		this.baseUrl = baseUrl

		this.i18n = i18n

		// create calendar with the user's time zone
		def jiraServiceContext = new JiraServiceContextImpl(user)
		def timeZone = timeZoneService.getUserTimeZoneInfo(jiraServiceContext).toTimeZone()
		calendar = new GregorianCalendar(timeZone)
	}

	/**
	 * Returns whether the passed field ID is related to a custom field (true) or a system field (false).
	 */
	private static boolean isCustomField(String fieldId) {
		fieldId.startsWith("customfield_")
	}

	private static boolean isCustomField(field) {
		isCustomField(field.id)
	}

	/**
	 * Returns whether the passed field name is related to a calculated field.
	 */
	private static boolean isCalculatedField(String customFieldName) {
		customFieldName in CALCULATED_FIELD_NAMES
	}

	private static boolean isCalculatedField(customField) {
		isCalculatedField(customField.name)
	}

	/**
	 * Returns the system-, custom- or calculated field value for the issue.
	 * Note: the returned value is strongly typed (String, Date, etc.).
	 *
	 * @param fieldId is either the ID of a system field (e.g. "assignee"),
	 * 				ID of a custom field (e.g. "customfield_10123"),
	 * 				or the ID of a calculated field (e.g. "Excel parent key").
	 * @param columnLayoutItem (optional) should be passed when the value is
	 *				to be obtained in an Issue Navigator context.
	 */
	public def getFieldValue(issue, String fieldId, columnLayoutItem = null) {
		if(isCalculatedField(fieldId)) {
			return getCalculatedFieldValue(issue, fieldId)
		}
		if(!isCustomField(fieldId)) {
			return getSystemFieldValue(issue, fieldId)
		}

		def customField = customFieldManager.getCustomFieldObject(fieldId)
		if(customField == null) {
			return "${fieldId} not found"
		}

		if(isCalculatedField(customField)) {
			return getCalculatedFieldValue(issue, customField)
		}

		return getCustomFieldValue(issue, customField, columnLayoutItem)
	}

	public def getFieldValue(issue, field, columnLayoutItem = null) {
		getFieldValue(issue, field.id, columnLayoutItem)
	}

	/**
	 * This convenience method is only to be used if you want to get a field value by the field name.
	 * The #getFieldValue(issue, String fieldId, columnLayoutItem = null) method is generally a safer option, because that accepts the immutable field ID (e.g. "customfield_10123"), not the name.
	 * This is more convenient because it can work out of the box without configuring the custom field IDs in the templates, but breaks when the custom field is renamed.
	 */
	public def getFieldValueByName(issue, String fieldName, columnLayoutItem = null) {
		def customField = customFieldManager.getCustomFieldObjectByName(fieldName)
		if(customField == null) {
			return "\"${fieldName}\" not found"
		}

		return getCustomFieldValue(issue, customField, columnLayoutItem)
	}

	/**
	 * This convenience method is similar to the previous one, but tolerates custom field renamings by receiving multiple name variants.
	 *
	 * @param fieldNameVariants is an array with multiple custom field names.
	 * 				The value of the first existing custom field name will be returned.
	 */
	public def getFieldValueByName(issue, String[] fieldNameVariants, columnLayoutItem = null) {
		def customField = null

		for(fieldName in fieldNameVariants) {
			customField = customFieldManager.getCustomFieldObjectByName(fieldName)
			if(customField) {
				break
			}
		}
		if(customField == null) {
			return "${fieldNameVariants} not found"
		}

		return getCustomFieldValue(issue, customField, columnLayoutItem)
	}


	private def getSystemFieldValue(issue, fieldId) {
		switch(fieldId) {
			case "project":
				issue.projectObject.name
				break

			case "versions":
				issue.affectedVersions.collect { it.name }.join(", ")
				break

			case "assignee":
				nullSafeUser(issue.assignee)
				break

			case "components":
				issue.components.collect { it.name }.join(", ")
				break

			case "comment":
				"${fieldId} not supported yet"
				break

			case "description":
				issue.description
				break

			case "duedate":
				nullSafeDate(issue.dueDate)
				break

			case "environment":
				issue.environment
				break

			case "fixVersions":
				issue.fixVersions.collect { it.name }.join(", ")
				break

			case "issuekey":
				issueToLink(issue)
				break

			case "number":
				"${fieldId} not supported yet"
				break

			case "issuetype":
				issue.issueTypeObject.nameTranslation
				break

			case "thumbnail":
				"${fieldId} not supported yet"
				break

			case "issuelinks":
				def linkCollection = issueViewUtil.getLinkCollection(issue, user)
				linkCollection.allIssues.collect { it.key }.join(", ")
				break

			case "lastViewed":
				"${fieldId} not supported yet"
				break

			case "workratio":
				def percentage = workRatio.getWorkRatio(issue)
				(percentage >= 0) ? (percentage / 100.0) : ""
				break

			case "subtasks":
				issue.subTaskObjects.collect { it.key }.join(", ")
				break

			case "attachment":
				"${fieldId} not supported yet"
				break

			case "priority":
				issue.priorityObject.nameTranslation
				break

			case "reporter":
				nullSafeUser(issue.reporter)
				break

			case "creator":
				nullSafeUser(issue.creator)
				break

			case "security":
				issue.securityLevel?.name
				break

			case "summary":
				summaryAsLink ? "${issue.summary}|${baseUrl}/browse/${issue.key}" : issue.summary
				break

			case "timetracking":
				"${fieldId} not supported yet"
				break

			case "created":
				nullSafeDateTime(issue.getCreated())
				break

			case "updated":
				nullSafeDateTime(issue.updated)
				break

			case "resolutiondate":
				nullSafeDateTime(issue.resolutionDate)
				break

			case "status":
				issue.statusObject.nameTranslation
				break

			case "resolution":
				issue.resolutionObject?.name
				break

			case "labels":
				issue.labels.join(", ")
				break

			case "worklog":
				"${fieldId} not supported yet"
				break

			case "timeoriginalestimate":
				secondsToHours(issue.originalEstimate)
				break

			case "timeestimate":
				secondsToHours(issue.estimate)
				break

			case "timespent":
				secondsToHours(issue.timeSpent)
				break

			case "aggregatetimespent":
				secondsToHours(issue.aggregateTimeSpent)
				break

			case "aggregatetimeestimate":
				secondsToHours(issue.aggregateRemainingEstimate)
				break

			case "aggregatetimeoriginalestimate":
				secondsToHours(issue.aggregateOriginalEstimate)
				break

			case "aggregateprogress":
				def percentage = AbstractProgressBarSystemField.calculateProgressPercentage(issue.aggregateTimeSpent, issue.aggregateRemainingEstimate)
				(percentage >= 0) ? (percentage / 100.0) : ""
				break

			case "progress":
				def percentage = AbstractProgressBarSystemField.calculateProgressPercentage(issue.timeSpent, issue.estimate)
				(percentage >= 0) ? (percentage / 100.0) : ""
				break

			case "votes":
				issue.votes
				break

			case "voter":
				"${fieldId} not supported yet"
				break

			case "watches":
				issue.watches
				break

			case "watcher":
				"${fieldId} not supported yet"
				break

			default:
				"${fieldId} not supported yet"
				break
		}
	}

	private def getCustomFieldValue(issue, customField, columnLayoutItem) {
		def customFieldTypeKey = customField.customFieldType.key
		def value = issue.getCustomFieldValue(customField)

		switch(customFieldTypeKey) {
			case "com.atlassian.jira.plugin.system.customfieldtypes:multicheckboxes":
			case "com.atlassian.jira.plugin.system.customfieldtypes:multiselect":
				(value != null) ? value.join(", ") : null
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:datepicker":
				nullSafeDate(value)
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:datetime":
				nullSafeDateTime(value)
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:multigrouppicker":
			case "com.atlassian.jira.plugin.system.customfieldtypes:grouppicker":
			case "com.atlassian.jira.plugin.system.customfieldtypes:multiversion":
			case "com.atlassian.jira.plugin.system.customfieldtypes:version":
				(value != null) ? value.collect { it.name }.join(", ") : null
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:labels":
				(value != null) ? value.collect { it.label }.join(", ") : null
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:float":
				value // to keep its original Double type (not to convert it to String)
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:project":
				value?.name
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:radiobuttons":
				value?.value
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:cascadingselect":
				(value != null) ? value.get(null).toString() + ((value.get("1") != null) ? " - " + value.get("1").toString() : "") : null
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:select":
				value?.value
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:textarea":
				value
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:textfield":
				value
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:url":
				value
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:multiuserpicker":
				(value != null) ? value.collect { nullSafeUser(it) }.join(", ") : null
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:userpicker":
				nullSafeUser(value)
				break

			// Jira Agile custom field types
			case "com.pyxis.greenhopper.jira:gh-sprint":
				(value != null) ? value.collect { it.name }.join(", ") : null
				break

			case "com.pyxis.greenhopper.jira:gh-epic-color":
				value
				break

			case "com.pyxis.greenhopper.jira:gh-epic-link":
				customField.customFieldType.getEpicDisplayName(value)
				break

			case "com.pyxis.greenhopper.jira:gh-epic-label":
				value
				break

			case "com.pyxis.greenhopper.jira:gh-epic-status":
				value
				break

			case "com.pyxis.greenhopper.jira:gh-lexo-rank":
				"${customFieldTypeKey} not supported yet"
				break

			// Jira Service Desk custom field types
			case "com.atlassian.servicedesk:sd-customer-organizations":
				(value != null) ? value.collect { it.name }.join(", ") : null
				break

			case "com.atlassian.servicedesk:sd-request-feedback":
				def score = value?.rate // get value.scale for the max value
				(score && sdExportCustomerSatisfactionComment) ? "${score}\n${getCustomerSatisfactionComment(issue)}" : score
				break

			case "com.atlassian.servicedesk:sd-request-participants":
				(value != null) ? value.collect { nullSafeUser(it) }.join(", ") : null
				break

			case "com.atlassian.servicedesk:sd-sla-field":
				(value != null) ? slaToHours(issue, customField, columnLayoutItem) : null
				break

			case "com.atlassian.servicedesk:vp-origin":
				customField.customFieldType.getDisplayedValue(value)?.name
				break

			// Jira Development custom field types
			case "com.atlassian.jira.plugins.jira-development-integration-plugin:devsummary":
				((value != null) && value.containsKey("summaryBean")) ? developmentSummaryToString(value["summaryBean"].summary) : null
				break

			// Color Custom Fields custom field types
			case "ru.andreymarkelov.atlas.plugins.colorfields:picker-color-custom-field":
			case "ru.andreymarkelov.atlas.plugins.colorfields:select-color-custom-field":
				value
				break

			// Git Integration custom field types
			case "com.xiplink.jira.git.jira_git_plugin:gitbranches":
				value?.join(", ")
				break

			case "com.xiplink.jira.git.jira_git_plugin:gitcommitsreferenced":
				value?.toString()
				break

			// Jira Misc Custom Fields (JMCF) custom field types
			case "com.innovalog.jmcf.jira-misc-custom-fields:calculated-date-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:last-field-change-datetime-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:statusdatetime":
			case "com.innovalog.jmcf.jira-misc-custom-fields:transitiondatetime":
				nullSafeDateTime(value)
				break

			case "com.innovalog.jmcf.jira-misc-custom-fields:calculated-duration-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:time-in-status-field":
				secondsToHours(value)
				break

			case "com.innovalog.jmcf.jira-misc-custom-fields:calculated-multi-select-field":
				(value != null) ? value.join(", ") : null
				break

			case "com.innovalog.jmcf.jira-misc-custom-fields:calculated-multi-user-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:transitioncallers":
				(value != null) ? value.collect{ nullSafeUser(it) }.join(", ") : null
				break

			case "com.innovalog.jmcf.jira-misc-custom-fields:calculated-number-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:parent-status-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:transition-count-field":
				value
				break

			case "com.innovalog.jmcf.jira-misc-custom-fields:calculated-single-select-field":
				(value != null) ? value.value : null
				break

			case "com.innovalog.jmcf.jira-misc-custom-fields:calculated-text-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:calculated-wiki-text-field":
				(value != null) ? xmlToDelimitedString(value, ", ") : null
				break

			case "com.innovalog.jmcf.jira-misc-custom-fields:calculated-user-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:last-field-change-author-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:statuscaller":
			case "com.innovalog.jmcf.jira-misc-custom-fields:transitioncaller":
				nullSafeUser(value)
				break

			// Insight custom field types
			case "com.riadalabs.jira.plugins.insight:rlabs-customfield-default-object":
			case "com.riadalabs.jira.plugins.insight:rlabs-customfield-object":
			case "com.riadalabs.jira.plugins.insight:rlabs-customfield-object-reference":
			case "com.riadalabs.jira.plugins.insight:rlabs-customfield-object-readonly":
			case "com.riadalabs.jira.plugins.insight:rlabs-customfield-object-multi":
			case "com.riadalabs.jira.plugins.insight:rlabs-customfield-object-reference-multi":
				(value != null) ? value.collect { insightObjectToString(it) }.sort { a, b -> a.toLowerCase() <=> b.toLowerCase() }.join(", ") : null
				break

			// nFeed custom field types
			case "com.valiantys.jira.plugins.SQLFeed:nfeed-standard-customfield-type":
			case "com.valiantys.jira.plugins.SQLFeed:com.valiantys.jira.plugins.sqlfeed.user.customfield.type":
			case "com.valiantys.jira.plugins.SQLFeed:com.valiantys.jira.plugins.sqlfeed.customfield.type": // deprecated field type
				(value != null) ? xmlToDelimitedString(nFeedDisplayToString(issue, customField), ", ") : null
				break

			case "com.valiantys.jira.plugins.SQLFeed:nfeed-date-customfield-type":
				nullSafeDate(value)
				break

			case "com.valiantys.jira.plugins.SQLFeed:nfeed-datetime-customfield-type":
				nullSafeDateTime(value)
				break

			// Portfolio custom field types
			case "com.atlassian.jpo:jpo-custom-field-baseline-start":
			case "com.atlassian.jpo:jpo-custom-field-baseline-end":
				nullSafeDate(value)
				break

			case "com.atlassian.jpo:jpo-custom-field-original-story-points":
				value
				break

			case "com.atlassian.jpo:jpo-custom-field-parent":
				(value != null) ? issueToLink(value, true) : null
				break

			case "com.atlassian.teams:rm-teams-custom-field-team":
				(value != null) ? value.description.title : null
				break

			// Profields custom field types
			case "com.deiser.jira.profields:profields.customfield.type.date":
			case "com.deiser.jira.profields:profieldscustomfielddate": // for pre-5.5.5 versions
				nullSafeDate(value)
				break

			case "com.deiser.jira.profields:profields.customfield.type.duration":
				(value != null) ? secondsToHours(value) : null
				break

			case "com.deiser.jira.profields:profields.customfield.type.group":
			case "com.deiser.jira.profields:profields.customfield.type.project":
			case "com.deiser.jira.profields:profieldscustomfieldprojectpicker": // for pre-5.5.5 versions
				value?.name
				break

			case "com.deiser.jira.profields:profields.customfield.type.group.multi":
			case "com.deiser.jira.profields:profields.customfield.type.project.multi":
				value?.collect { it?.name }?.join(", ")
				break

			case "com.deiser.jira.profields:profields.customfield.type.list":
			case "com.deiser.jira.profields:profields.customfield.type.priority":
			case "com.deiser.jira.profields:profields.customfield.type.status":
				value?.text
				break

			case "com.deiser.jira.profields:profields.customfield.type.list.multi":
				value?.collect { it?.text }?.join(", ")
				break

			case "com.deiser.jira.profields:profieldscustomfieldlist": // for pre-5.5.5 versions
				value?.collect { it?.value }?.join(", ")
				break

			case "com.deiser.jira.profields:profields.customfield.type.number":
			case "com.deiser.jira.profields:profields.customfield.type.text":
			case "com.deiser.jira.profields:profieldscustomfieldnumber": // for pre-5.5.5 versions
			case "com.deiser.jira.profields:profieldscustomfieldtext": // for pre-5.5.5 versions
			case "com.deiser.jira.profields:profieldscustomfieldoriginalestimate": // for pre-5.5.5 versions
				value
				break

			case "com.deiser.jira.profields:profields.customfield.type.user":
			case "com.deiser.jira.profields:profieldscustomfielduserpicker": // for pre-5.5.5 versions
				nullSafeUser(value)
				break

			case "com.deiser.jira.profields:profields.customfield.type.user.multi":
				value?.collect { nullSafeUser(it) }?.join(", ")
				break

			// ScriptRunner custom field types
			case "com.onresolve.jira.groovy.groovyrunner:single-issue-picker-cf":
				(value != null) ? issueToLink(value) : null
				break

			case "com.onresolve.jira.groovy.groovyrunner:single-remote-issue-picker-cf":
				value?.key
				break

			case "com.onresolve.jira.groovy.groovyrunner:multiple-issue-picker-cf":
			case "com.onresolve.jira.groovy.groovyrunner:multiple-remote-issue-picker-cf":
				value?.collect { it.key }?.join(", ")
				break

			case "com.onresolve.jira.groovy.groovyrunner:scripted-field":
				// export based on the searcher or as string if there is custom template set
				switch(getSearcherKeyOrBlank(customField)) {
					case "com.onresolve.jira.groovy.groovyrunner:datetimerange":
						nullSafeDateTime(value)
						break

					case "com.onresolve.jira.groovy.groovyrunner:durationsearcher":
						(value != null) ? secondsToHours(value) : null
						break

					case "com.onresolve.jira.groovy.groovyrunner:exactnumber":
						value
						break

					case "com.onresolve.jira.groovy.groovyrunner:grouppickersearcher":
					case "com.onresolve.jira.groovy.groovyrunner:projectsearcher":
						value?.name
						break

					case "com.onresolve.jira.groovy.groovyrunner:issuekeysearcher":
						(value != null) ? issueToLink(value, true) : null
						break

					case "com.onresolve.jira.groovy.groovyrunner:multigrouppickersearcher":
					case "com.onresolve.jira.groovy.groovyrunner:versionsearcher":
						value?.collect { it?.name }?.join(", ")
						break

					case "com.onresolve.jira.groovy.groovyrunner:multiuserpickersearcher":
						value?.collect { nullSafeUser(it) }?.join(", ")
						break

					case "com.onresolve.jira.groovy.groovyrunner:userpickersearcher":
						nullSafeUser(value)
						break

					default:
						// other, including "com.onresolve.jira.groovy.groovyrunner:textsearcher" and script fields with custom templates
						getDefaultCustomFieldValueAsText(issue, customField, columnLayoutItem, value)
						break
				}
				break

			case "com.onresolve.jira.groovy.groovyrunner:single-database-picker-cf":
			case "com.onresolve.jira.groovy.groovyrunner:multiple-database-picker-cf":
				getDefaultCustomFieldValueAsText(issue, customField, columnLayoutItem, value)
				break

			// Tempo custom field types
			case "com.tempoplugin.tempo-accounts:accounts.customfield":
				customField.customFieldType.getChangelogString(customField, value)
				break

			case "com.tempoplugin.tempo-plan-core:tp.iteration.customfield":
				// due to a Tempo bug, the rendered HTML value is broken (no quotes around the value of a href attribute) and we cannot parse it
				// we must fall back to the value object's toString() method, thus the rendered output will be a numeric ID value
				(value != null) ? getFieldValueAsText(value) : null
				break

			case "com.tempoplugin.tempo-teams:team.customfield":
				(value != null) ? getFieldValueAsText(issue, customField, columnLayoutItem) : null
				break

			// Time to SLA custom field types
			case "plugin.tts:overduestatusDbVersion":
				(value != null) ? overdueStatusToString(value) : null
				break

			case "plugin.tts:slaindicatorDbVersion":
				(value != null) ? slaIndicatorToString(value) : null
				break

			case "plugin.tts:slaoverviewDbVersion-cf":
				(value != null) ? slaOverviewToString(value) : null
				break

			case "plugin.tts:timetoSlaDbVersion":
				(value != null) ? timeToSlaToString(value) : null
				break

			case "plugin.tts:tts-date-cf":
				nullSafeDateTime(value)
				break

			case "plugin.tts:ttsText-cf":
				value
				break

			// Xray custom field types
			case "com.xpandit.plugins.xray:automated-test-type-custom-field":
			case "com.xpandit.plugins.xray:path-editor-custom-field":
			case "com.xpandit.plugins.xray:precondition-editor-custom-field":
			case "com.xpandit.plugins.xray:precondition-test-type-custom-field":
			case "com.xpandit.plugins.xray:steps-editor-custom-field":
			case "com.xpandit.plugins.xray:test-repository-path-custom-field":
			case "com.xpandit.plugins.xray:test-run-status-custom-field":
			case "com.xpandit.plugins.xray:test-type-custom-field":
				value
				break

			case "com.xpandit.plugins.xray:issue-link-count-custom-field":
			case "com.xpandit.plugins.xray:manual-test-steps-count-count-custom-field":
				value?.toInteger()
				break

			case "com.xpandit.plugins.xray:manual-test-steps-custom-field":
				value ? xrayManualTestStepsToString(value) : null
				break

			case "com.xpandit.plugins.xray:manual-test-steps-export-custom-field":
				value ? xmlToDelimitedString(value, ", ") : null
				break

			case "com.xpandit.plugins.xray:precondition-test-custom-field":
			case "com.xpandit.plugins.xray:test-plan-custom-field":
			case "com.xpandit.plugins.xray:test-plans-associated-with-test-custom-field":
			case "com.xpandit.plugins.xray:test-precondition-custom-field":
			case "com.xpandit.plugins.xray:test-sets-custom-field":
			case "com.xpandit.plugins.xray:test-sets-tests-custom-field":
			case "com.xpandit.plugins.xray:tests-associated-with-test-plan-custom-field":
				value ? value.collect { xrayIssueKeyValueToString(it) }.join(", ") : null
				break

			case "com.xpandit.plugins.xray:requirement-group-test-status-custom-field":
				value ? value.collect { it.status.label }.join(", ") : null
				break

			case "com.xpandit.plugins.xray:requirement-status-custom-field":
				value ? value.collect { it.status }.join(", ") : null
				break

			case "com.xpandit.plugins.xray:requirement-version-test-status-custom-field":
			case "com.xpandit.plugins.xray:test-environments-custom-field":
				value ? value.collect { it.label }.join(", ") : null
				break

			case "com.xpandit.plugins.xray:testexec-tests-custom-field":
				value ? value.collect { xrayIssueKeyValueToString(!it.metaClass.respondsTo(it, "a") ? it : it.a()) }.join(", ") : null
				break

			case "com.xpandit.plugins.xray:test-execution-defects-custom-field":
				// when the issue is not a Test Execution, value is an empty list
				// when the issue is a Test Execution with no defects, value is null
				// when the issue is a Test Execution with defects, value is a non-empty list
				(value != null) ? (value.empty ? null : value.size()) : 0
				break

			case "com.xpandit.plugins.xray:test-execution-status-custom-field":
			case "com.xpandit.plugins.xray:test-plan-status-custom-field":
			case "com.xpandit.plugins.xray:test-set-status-custom-field":
				if (value != null) {
					if (value.metaClass.respondsTo(value, "getStatuses")) {
						value.statuses.collect { it.name + " : " + it.statusCount }.join(", ")
					} else if (!value.metaClass.respondsTo(value, "getTestStatuses") || value.testStatuses) {
						value.toString().replaceAll(/\([\d.%]+\)/, "")
					}
				} else {
					null
				}
				break

			default:
				getDefaultCustomFieldValueAsText(issue, customField, columnLayoutItem, value)
				break
		}
	}

	private def getDefaultCustomFieldValueAsText(issue, customField, columnLayoutItem, value) {
		(columnLayoutItem != null) ? getFieldValueAsText(issue, customField, columnLayoutItem) : getFieldValueAsText(value)
	}

	private def getCalculatedFieldValue(issue, customField) {
		return getCalculatedFieldValue(issue, customField.name)
	}

	private def getCalculatedFieldValue(issue, String fieldName) {
		def comments = commentManager.getCommentsForUser(issue, user)

		switch(fieldName) {
			case EXCEL_PARENT_KEY_FIELD_NAME:
				issue.isSubTask() ? issueToLink(issue.parentObject) : ""
				break

			case EXCEL_COMMENTS_CALCULATED_FIELD_NAME:
				!comments.empty ? comments.collect { commentToString(it) }.join("\n\n") : ""
				break

			case EXCEL_FIRST_COMMENT_CALCULATED_FIELD_NAME:
				!comments.empty ? commentToString(comments.first()) : ""
				break

			case EXCEL_FIRST_COMMENT_TEXT_CALCULATED_FIELD_NAME:
				!comments.empty ? comments.first().body : ""
				break

			case EXCEL_FIRST_COMMENT_AUTHOR_CALCULATED_FIELD_NAME:
				!comments.empty ? comments.first().authorFullName : ""
				break

			case EXCEL_FIRST_COMMENT_DATE_CALCULATED_FIELD_NAME:
				!comments.empty ? nullSafeDateTime(comments.first().created) : ""
				break

			case EXCEL_LAST_COMMENT_CALCULATED_FIELD_NAME:
				!comments.empty ? commentToString(comments.last()) : ""
				break

			case EXCEL_LAST_COMMENT_TEXT_CALCULATED_FIELD_NAME:
				!comments.empty ? comments.last().body : ""
				break

			case EXCEL_LAST_COMMENT_AUTHOR_CALCULATED_FIELD_NAME:
				!comments.empty ? comments.last().authorFullName : ""
				break

			case EXCEL_LAST_COMMENT_DATE_CALCULATED_FIELD_NAME:
				!comments.empty ? nullSafeDateTime(comments.last().created) : ""
				break

			default:
				"${fieldName} is an undefined calculated field"
				break
		}
	}

	/**
	 * Returns the Excel cell type for the issue field.
	 * It is one of "number", "date" or "string".
	 */
	public def getFieldCellType(String fieldId) {
		if(isCalculatedField(fieldId)) {
			return getCalculatedFieldCellType(issue, fieldId)
		}
		if(!isCustomField(fieldId)) {
			return getSystemFieldCellType(fieldId)
		}

		def customField = customFieldManager.getCustomFieldObject(fieldId)
		if(customField == null)  {
			return "string"
		}

		if(isCalculatedField(customField)) {
			return getCalculatedFieldCellType(customField)
		}

		return getCustomFieldCellType(customField)
	}

	public def getFieldCellType(field) {
		getFieldCellType(field.id)
	}

	private def getSystemFieldCellType(fieldId) {
		switch(fieldId) {
			case "workratio":
			case "timeoriginalestimate":
			case "timeestimate":
			case "timespent":
			case "aggregatetimespent":
			case "aggregatetimeestimate":
			case "aggregatetimeoriginalestimate":
			case "aggregateprogress":
			case "progress":
			case "votes":
			case "watches":
				"number"
				break

			case "duedate":
			case "created":
			case "updated":
			case "resolutiondate":
				"date"
				break

			case "issuekey":
				issueKeyAsLink ? "link" : "string"
				break

			case "summary":
				summaryAsLink ? "link" : "string"
				break

			default:
				"string"
		}
	}

	private def getCustomFieldCellType(customField) {
		switch(customField.customFieldType.key) {
			case "com.atlassian.jira.plugin.system.customfieldtypes:datepicker":
			case "com.atlassian.jira.plugin.system.customfieldtypes:datetime":
				"date"
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:float":
				"number"
				break

			// Jira Service Desk custom field types
			case "com.atlassian.servicedesk:sd-request-feedback":
				!sdExportCustomerSatisfactionComment ? "number" : "string"
				break

			case "com.atlassian.servicedesk:sd-sla-field":
				"number"
				break

			// Jira Misc Custom Fields (JMCF) custom field types
			case "com.innovalog.jmcf.jira-misc-custom-fields:calculated-date-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:last-field-change-datetime-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:statusdatetime":
			case "com.innovalog.jmcf.jira-misc-custom-fields:transitiondatetime":
				"date"
				break

			case "com.innovalog.jmcf.jira-misc-custom-fields:calculated-duration-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:calculated-number-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:time-in-status-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:transition-count-field":
				"number"
				break

			// nFeed custom field types
			case "com.valiantys.jira.plugins.SQLFeed:nfeed-date-customfield-type":
			case "com.valiantys.jira.plugins.SQLFeed:nfeed-datetime-customfield-type":
				"date"
				break

			// Portfolio custom field types
			case "com.atlassian.jpo:jpo-custom-field-baseline-start":
			case "com.atlassian.jpo:jpo-custom-field-baseline-end":
				"date"
				break

			case "com.atlassian.jpo:jpo-custom-field-original-story-points":
				"number"
				break

			case "com.atlassian.jpo:jpo-custom-field-parent":
				issueKeyAsLink ? "link" : "string"
				break

			// Profields custom field types
			case "com.deiser.jira.profields:profields.customfield.type.date":
			case "com.deiser.jira.profields:profieldscustomfielddate": // for pre-5.5.5 versions
				"date"
				break

			case "com.deiser.jira.profields:profields.customfield.type.duration":
			case "com.deiser.jira.profields:profields.customfield.type.number":
			case "com.deiser.jira.profields:profieldscustomfieldnumber": // for pre-5.5.5 versions
				"number"
				break

			// ScriptRunner custom field types
			case "com.onresolve.jira.groovy.groovyrunner:single-issue-picker-cf":
				issueKeyAsLink ? "link" : "string"
				break

			case "com.onresolve.jira.groovy.groovyrunner:single-remote-issue-picker-cf":
			case "com.onresolve.jira.groovy.groovyrunner:multiple-issue-picker-cf":
			case "com.onresolve.jira.groovy.groovyrunner:multiple-remote-issue-picker-cf":
				"string"
				break

			case "com.onresolve.jira.groovy.groovyrunner:scripted-field":
				switch(getSearcherKeyOrBlank(customField)) {
					case "com.onresolve.jira.groovy.groovyrunner:datetimerange":
						"date"
						break

					case "com.onresolve.jira.groovy.groovyrunner:durationsearcher":
					case "com.onresolve.jira.groovy.groovyrunner:exactnumber":
						"number"
						break

					case "com.onresolve.jira.groovy.groovyrunner:issuekeysearcher":
						issueKeyAsLink ? "link" : "string"
						break

					default:
						"string"
						break
				}
				break

			// Time to SLA custom field types
			case "plugin.tts:tts-date-cf":
				"date"
				break

			// Xray custom field types
			case "com.xpandit.plugins.xray:issue-link-count-custom-field":
			case "com.xpandit.plugins.xray:manual-test-steps-count-count-custom-field":
			case "com.xpandit.plugins.xray:test-execution-defects-custom-field":
				"number"
				break

			default:
				"string"
		}
	}

	private def getCalculatedFieldCellType(customField) {
		return getCalculatedFieldCellType(customField.name)
	}

	private def getCalculatedFieldCellType(String fieldName) {
		switch(fieldName) {
			case EXCEL_PARENT_KEY_FIELD_NAME:
				issueKeyAsLink ? "link" : "string"
				break

			case EXCEL_FIRST_COMMENT_DATE_CALCULATED_FIELD_NAME:
			case EXCEL_LAST_COMMENT_DATE_CALCULATED_FIELD_NAME:
				"date"
				break

			default:
				"string"
		}
	}

	/**
	 * Returns the format specifier string (ex: "dd/mmm/yyyy") for the issue field.
	 * The format specifier is typically used for number- or date type fields.
	 */
	public def getFieldFormat(String fieldId) {
		if(isCalculatedField(fieldId)) {
			return getCalculatedFieldFormat(issue, fieldId)
		}
		if(!isCustomField(fieldId)) {
			return getSystemFieldFormat(fieldId)
		}

		def customField = customFieldManager.getCustomFieldObject(fieldId)
		if(customField == null) {
			return null
		}

		if(isCalculatedField(customField)) {
			return getCalculatedFieldFormat(customField)
		}

		return getCustomFieldFormat(customField)
	}

	public def getFieldFormat(field) {
		getFieldFormat(field.id)
	}

	private def getSystemFieldFormat(fieldId) {
		switch(fieldId) {
			case "workratio":
			case "progress":
			case "aggregateprogress":
				"0.00%"
				break

			case "timeoriginalestimate":
			case "timeestimate":
			case "timespent":
			case "aggregatetimespent":
			case "aggregatetimeestimate":
			case "aggregatetimeoriginalestimate":
				"0.00"
				break

			case "duedate":
			case "created":
			case "updated":
			case "resolutiondate":
				"dd/mmm/yyyy"
				break

			default:
				// no custom format
				null
		}
	}

	private def getCustomFieldFormat(customField) {
		switch(customField.customFieldType.key) {
			case "com.atlassian.jira.plugin.system.customfieldtypes:datepicker":
				"dd/mmm/yyyy"
				break

			case "com.atlassian.jira.plugin.system.customfieldtypes:datetime":
				"dd/mmm/yyyy hh:mm"
				break

			// Jira Service Desk custom field types
			case "com.atlassian.servicedesk:sd-sla-field":
				"0.00\" h\";[RED]-0.00\" h\""
				break

			// Jira Misc Custom Fields (JMCF) custom field types
			case "com.innovalog.jmcf.jira-misc-custom-fields:calculated-date-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:last-field-change-datetime-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:statusdatetime":
			case "com.innovalog.jmcf.jira-misc-custom-fields:transitiondatetime":
				"dd/mmm/yyyy hh:mm"
				break

			case "com.innovalog.jmcf.jira-misc-custom-fields:calculated-duration-field":
			case "com.innovalog.jmcf.jira-misc-custom-fields:time-in-status-field":
				"0.00"
				break

			// nFeed custom field types
			case "com.valiantys.jira.plugins.SQLFeed:nfeed-date-customfield-type":
				"dd/mmm/yyyy"
				break

			case "com.valiantys.jira.plugins.SQLFeed:nfeed-datetime-customfield-type":
				"dd/mmm/yyyy hh:mm"
				break

			// Profields custom field types
			case "com.deiser.jira.profields:profields.customfield.type.duration":
				"0.00"
				break

			// ScriptRunner custom field types
			case "com.onresolve.jira.groovy.groovyrunner:scripted-field":
				switch (getSearcherKeyOrBlank(customField)) {
					case "com.onresolve.jira.groovy.groovyrunner:datetimerange":
						"dd/mmm/yyyy hh:mm"
						break

					case "com.onresolve.jira.groovy.groovyrunner:durationsearcher":
						"0.00"
						break

					default:
						null
				}
				break

			// Time to SLA custom field types
			case "plugin.tts:tts-date-cf":
				"dd/mmm/yyyy hh:mm"
				break

			default:
				// no custom format
				null
		}
	}

	private def getCalculatedFieldFormat(customField) {
		return getCalculatedFieldFormat(customField.name);
	}

	private def getCalculatedFieldFormat(String fieldName) {
		switch(fieldName) {
			case EXCEL_FIRST_COMMENT_DATE_CALCULATED_FIELD_NAME:
			case EXCEL_LAST_COMMENT_DATE_CALCULATED_FIELD_NAME:
				"dd/mmm/yyyy hh:mm"
				break

			default:
				// no custom format
				null
		}
	}

	/**
	 * Returns the raw rendered HTML value for fields without an explicit value formatter.
	 */
	private def getFieldValueAsHtml(issue, columnLayoutItem) {
		columnLayoutItem.getHtml([:], issue).replace("&nbsp;", " ")
	}

	/**
	 * Returns the raw text value for fields without an explicit value formatter.
	 * (This method relies on parsing the text parts from the HTML value.)
	 */
	private def getFieldValueAsText(issue, field, columnLayoutItem) {
		// create a dummy column layout item if not passed (necessary for the HTML rendering)
		if(columnLayoutItem == null) {
			columnLayoutItem = new ColumnLayoutItemImpl(field, 1)
		}

		// render HTML
		def valueHtml = getFieldValueAsHtml(issue, columnLayoutItem)

		// get text value from the HTML value
		def value = !valueHtml.isAllWhitespace() ? xmlToDelimitedString(valueHtml, " ") : null

		return value ?: valueHtml
	}

	/**
	 * Returns the value for textual fields without an explicit value formatter.
	 * This method relies on the toString() method.
	 */
	private def getFieldValueAsText(value) {
		value.toString().trim()
	}

	/**
	 * Returns the issue key (optionally with the summary) as plain text or as clickable link.
	 */
	private def issueToLink(issue, showSummary = false) {
		def text = issue.key + (showSummary ? " " + issue.summary : "")
		return issueKeyAsLink ? (text + "|${baseUrl}/browse/" + issue.key) : text
	}

	/**
	 * Returns the string that contains the comment text, the comment author and the comment date.
	 */
	private def commentToString(def comment) {
		def commentString = comment.body + "\n---"

		def commentUser = nullSafeUser(comment.authorApplicationUser)
		if(!"".equals(commentUser)) {
			commentString += " by " + commentUser
		}

		def commentCalendar = nullSafeDateTime(comment.created)
		if(!"".equals(commentCalendar)) {
			commentString += " at " + commentCalendar.format("dd/MMM/yyyy HH:mm")
		}

		return commentString
	}

	/**
	 * Should be used for the values without time component and time zone.
	 */
	private def nullSafeDate(def date) {
		(date != null) ? date : ""
	}

	/**
	 * Should be used for the values with time component and time zone.
	 */
	private def nullSafeDateTime(def date) {
		if (date != null) {
			calendar.setTime(date)
			return calendar
		} else {
			return ""
		}
	}

	private static def nullSafeUser(def user) {
		(user != null) ? user.displayName : ""
	}

	/**
	 * Returns the value or 0 if the value is <code>null</code>.
	 */
	private static def nullSafeNumber(def number) {
		(number != null) ? number : 0
	}

	private static Double secondsToHours(def seconds) {
		(seconds != null) ? (seconds / (60.0*60.0)) : null
	}

	// - Jira Service Desk integration ----------------------------------------

	private def slaToHours(def issue, def customField, def columnLayoutItem) {
		def goalDetailedViewDataList = issueIdToGoalDetailedViewDataList.get(issue.id)
		if (goalDetailedViewDataList == null) {
			goalDetailedViewDataList = customField.customFieldType?.goalViewService?.getGoalDataForDetailView(user, issue)
			issueIdToGoalDetailedViewDataList.put(issue.id, goalDetailedViewDataList)
		}

		def filteredGoalDetailedViewDataList = goalDetailedViewDataList?.findAll { it.metricName == customField.name }
		def goalDetailedViewData = (filteredGoalDetailedViewDataList) ? filteredGoalDetailedViewDataList.first() : null
		if (goalDetailedViewData && goalDetailedViewData.hasProperty("remainingTimeInMilliseconds")) {
			// modern approach
			return goalDetailedViewData.remainingTimeInMilliseconds / (60*60*1000)
		} else {
			// legacy approach
			def text = getFieldValueAsText(issue, customField, columnLayoutItem)
			if (!text.isAllWhitespace()) {
				def parts = text.split(':')
				if (parts.size() == 2) {
					def sign = (parts[0].indexOf('-') == -1) ? 1 : -1
					def hours = parts[0].replace("-","").toInteger()
					def minutes = parts[1].toInteger()
					return sign*(hours + minutes/60.0) // hours as decimals number (-1:45 is exported as "-1.75" and formatted as red "-1.75 h")
				}
			}
		}
		return null
	}

	private def getCustomerSatisfactionComment(def issue) {
		// lazily load dependency (attempt only once)
		if (!requestFeedbackInternalServiceInitialized) {
			try {
				requestFeedbackInternalService = ComponentAccessor.getOSGiComponentInstanceOfType(ComponentAccessor.pluginAccessor.classLoader.loadClass("com.atlassian.servicedesk.internal.api.feedback.RequestFeedbackInternalService")) // FIXME deprecated, use ClassHelper instead
			} catch (Exception ex) {
				log.warn("Failed to find the RequestFeedbackInternalService", ex)
			}
			requestFeedbackInternalServiceInitialized = true
		}
		requestFeedbackInternalService?.getFeedback(issue).getOrNull()?.getComment().getOrNull() ?: "No comment provided."
	}

	// - Jira Development integration -----------------------------------------

	private def developmentSummaryToString(def summary) {
		// see https://confluence.atlassian.com/jirasoftwareserver/viewing-the-development-information-for-an-issue-938845580.html
		def entityNames = [
				"branch" : ["branch", "branches"],
				"repository" : ["commit", "commits"], // misleading property name
				"pullrequest" : ["pull request", "pull requests"]
		]
		def result = []

		// branches
		def branchCount = summary.branch.overall.count
		if (branchCount) {
			result << (branchCount + " " + entityNames["branch"][(branchCount == 1) ? 0 : 1])
		}

		// commits
		def commitCount = summary.repository.overall.count
		if (commitCount) {
			result << (commitCount + " " + entityNames["repository"][(commitCount == 1) ? 0 : 1])
		}

		// pull requests
		def pullRequestOverAll = summary.pullrequest.overall
		def stateCount = pullRequestOverAll.stateCount
		if (stateCount) {
			if (pullRequestOverAll.hasProperty("details")) {
				def pullRequestStrings = []

				def openCount = pullRequestOverAll.details.openCount
				if (openCount) {
					pullRequestStrings << (openCount + " open")
				}
				def mergedCount = pullRequestOverAll.details.mergedCount
				if (mergedCount) {
					pullRequestStrings << (mergedCount + " merged")
				}
				def declinedCount = pullRequestOverAll.details.declinedCount
				if (declinedCount) {
					pullRequestStrings << (declinedCount + " declined")
				}

				def sumCount = openCount + mergedCount + declinedCount
				if (pullRequestStrings) {
					result << (pullRequestStrings.join(" + ") + " " + entityNames["pullrequest"][(sumCount == 1) ? 0 : 1])
				}
			} else {
				// details are not available in earlier Jira 7.x versions
				result << (stateCount + " " + entityNames["pullrequest"][(stateCount == 1) ? 0 : 1])
			}
		}

		return result.join(", ")
	}

	// - Insight integration --------------------------------------------------

	/**
	 * Returns the string that contains the Insight object's name, key (optionally) and attributes (optionally).
	 */
	private def insightObjectToString(def object) {
		def objectStringBuilder = StringBuilder.newInstance()

		if(insightObjectsWithNames) {
			objectStringBuilder << object.name
		}

		if(insightObjectsWithKeys) {
			if(objectStringBuilder.size() > 0) {
				objectStringBuilder << " "
			}
			objectStringBuilder << "[" << object.objectKey << "]"
		}

		if(insightObjectsWithAttributes) {
			// lazily load dependencies
			if(insightObjectFacade == null) {
				insightObjectFacade = ComponentAccessor.getOSGiComponentInstanceOfType(ComponentAccessor.pluginAccessor.classLoader.findClass("com.riadalabs.jira.plugins.insight.channel.external.api.facade.ObjectFacade")) // FIXME deprecated, use ClassHelper instead
			}
			if(insightObjectTypeAttributeFacade == null) {
				insightObjectTypeAttributeFacade = ComponentAccessor.getOSGiComponentInstanceOfType(ComponentAccessor.pluginAccessor.classLoader.findClass("com.riadalabs.jira.plugins.insight.channel.external.api.facade.ObjectTypeAttributeFacade")) // FIXME deprecated, use ClassHelper instead
			}

			def attributesString = object.objectAttributeBeans.collect { objectAttribute ->
				def nameAndValueStringBuilder = StringBuilder.newInstance()
				if(objectAttribute.id != null) {
					def attribute = insightObjectTypeAttributeFacade.loadObjectTypeAttributeBean(objectAttribute.objectTypeAttributeId)
					if(attribute != null) {
						nameAndValueStringBuilder << attribute.name
						nameAndValueStringBuilder << "="
						nameAndValueStringBuilder << objectAttribute.objectAttributeValueBeans.collect { attributeValue ->
							(attributeValue.referencedObjectBeanId == null) ? attributeValue.value.toString() : insightObjectFacade.loadObjectBean(attributeValue.referencedObjectBeanId)?.name
						}.join(";")
					}
				}
				return nameAndValueStringBuilder.toString()
			}.findAll().join(", ")

			if(objectStringBuilder.size() > 0) {
				objectStringBuilder << " "
			}
			objectStringBuilder << "(" << attributesString << ")"
		}

		return objectStringBuilder.toString()
	}

	// - nFeed integration ----------------------------------------------------

	private def nFeedDisplayToString(def issue, def customField) {
		// lazily load dependency (attempt only once)
		if (!nFeedInitialized) {
			try {
				nFeedFieldDisplayService = ClassHelper.getOSGiComponentInstanceOfType("com.valiantys.nfeed.api.IFieldDisplayService")
			} catch (Exception ex) {
				log.warn("Failed to find the nFeed API", ex)
			}
			nFeedInitialized = true
		}

		(nFeedFieldDisplayService != null) ? nFeedFieldDisplayService.getDisplayResult(issue.id, customField.id).getDisplay() : getFieldValueAsText(issue.getCustomFieldValue(customField))
	}

	/**
	 * Returns the body text fragments from the passed XML joined with the delimiter.
	 * The XML is not required to be well-formatted and can even be plain text.
	 */
	private def xmlToDelimitedString(def xml, def delimiter) {
		xml.split(/<.*?>/).collect { it.trim() }.findAll { !it.allWhitespace }.join(delimiter)
	}

	// - ScriptRunner integration ---------------------------------------------

	/**
	 * Returns the searcher key for the script field or a blank string if it uses a custom template.
	 */
	private def getSearcherKeyOrBlank(customField) {
		try {
			return !isScriptFieldWithCustomTemplate(customField) ? getSearcherKey(customField) : ""
		} catch (Exception ex) {
			log.warn("Failed to get the searcherKey or modelTemplate for ${customField.id}", ex)
			return ""
		}
	}

	/**
	 * Returns true if the custom field uses a custom template.
	 */
	private def isScriptFieldWithCustomTemplate(customField) {
		if (scriptFieldIdsWithCustomTemplate == null) {
			def configurations = getScriptFieldConfigurations()
			scriptFieldIdsWithCustomTemplate = (configurations instanceof List) ?
					// ScriptRunner 6 or later
					configurations
							.findAll { it.modelTemplate.modelName == "custom" }
							.collect { it.customFieldId }
					:
					// ScriptRunner 5 or earlier
					configurations
							.findAll { k, v -> (v.modelTemplate == "custom") && (!v.metaClass.respondsTo(v, "isCannedScript") || !v.isCannedScript()) } // second condition is required in modern ScriptRunner versions
							.collect { k, v -> v.hasProperty("customFieldId") ? v.customFieldId : v.id.toLong() } // property has been renamed
		}

		return scriptFieldIdsWithCustomTemplate.contains(customField.idAsLong)
	}

	private def getSearcherKey(customField) {
		def customFieldId = customField.idAsLong
		if (!fieldIdToSearcherKey.containsKey(customFieldId)) {
			fieldIdToSearcherKey[customFieldId] = customField.searcherRef.get().descriptor.completeKey
		}
		return fieldIdToSearcherKey[customFieldId]
	}

	/**
	 * Returns the field configurations as List or Map, depending on the ScriptRunner version.
	 */
	private def getScriptFieldConfigurations() {
		if (scriptFieldConfigurations == null) {
			def version
			try {
				version = ComponentAccessor.pluginAccessor.getPlugin("com.onresolve.jira.groovy.groovyrunner").pluginInformation.version
				if (VersionHelper.isLaterOrSame(version, "6")) {
					// ScriptRunner 6 or later
					def scriptRunnerInstance = ClassHelper.getOSGiComponentInstanceOfType("com.onresolve.scriptrunner.runner.ScriptRunner")
					def classToLoad = VersionHelper.isLaterOrSame(version, "6.6.0") ? "com.onresolve.scriptrunner.fields.ScriptFieldConfigurationService" : "com.onresolve.scriptrunner.fields.ScriptFieldManager"
					def clazz = ComponentAccessor.pluginAccessor.classLoader.loadClass(classToLoad)
					scriptFieldConfigurations = scriptRunnerInstance.getBean(clazz).load()
				} else {
					// ScriptRunner 5 or earlier
					def scriptFieldManager = ClassHelper.getOSGiComponentInstanceOfType("com.onresolve.scriptrunner.runner.ScriptFieldManager")
					scriptFieldConfigurations = scriptFieldManager.configs
				}
			} catch (Exception ex) {
				log.warn("Failed to load field configurations (ScriptRunner version: ${version})", ex)
				scriptFieldConfigurations = []
			}
		}

		return scriptFieldConfigurations
	}

	// - Time to SLA integration ----------------------------------------------

	/**
	 * Returns a structured text representation of an "Overdue Status" value ("Time to SLA" app).
	 */
	private def overdueStatusToString(def ttsFieldValue) {
		ttsFieldValue.collect {
			"[" + it.description + "]\n" +
			INDENTATION + it.overdue
		}.join("\n")
	}

	/**
	 * Returns a structured text representation of an "SLA Indicator" value ("Time to SLA" app).
	 */
	private def slaIndicatorToString(def ttsFieldValue) {
		ttsFieldValue.collect {
			it.result + "\n" +
			it.overdues.collect {
				"[" + it.value.slaName + "]\n" +
				INDENTATION + i18n.getText("tts.slaBox.durations.overdue") + ": " + it.value.breachString + "\n" +
				INDENTATION + i18n.getText("tts.report.slareport.expectedTargetDate") + ": " + ((it.value.expectedTargetDate) ? it.value.expectedTargetDate : "-") + "\n" +
				INDENTATION + i18n.getText("tts.report.slareport.actualTargetDate") + ": " + ((it.value.actualTargetDate) ? it.value.actualTargetDate : "-")
			}.join("\n")
		}.join("\n")
	}

	/**
	 * Returns a structured text representation of an "SLA Overview" value ("Time to SLA" app).
	 */
	private def slaOverviewToString(def ttsFieldValue) {
		ttsFieldValue.collect {
				"[" + it.slaName + "]\n" +
				INDENTATION + i18n.getText("slaoverviewtabpanel.header.slaStatus").trim() + ": " + i18n.getText("slaoverviewtabpanel." + it.status.name()) + "\n" + // FIXME trim() is needed to remove the extra space at the end of the text
				INDENTATION + i18n.getText("slaoverviewtabpanel.header.workingDuration") + ": " + it.workingDuration + "\n" +
				INDENTATION + i18n.getText("slaoverviewtabpanel.header.pauseDuration") + ": " + it.pausedDuration + "\n" +
				INDENTATION + i18n.getText("slaoverviewtabpanel.header.slaValue") + ": " + it.slaValueAsTimeString + "\n" +
				INDENTATION + i18n.getText("slaoverviewtabpanel.header.calendar") + ": " + it.workingCalendar
		}.join("\n")
	}

	/**
	 * Returns a structured text representation of an "Time to SLA" value ("Time to SLA" app).
	 */
	private def timeToSlaToString(def ttsFieldValue) {
		ttsFieldValue.collect {
			"[" + it.slaName + "]\n" +
			INDENTATION + getSlaStatusText(it) + "\n" +
			INDENTATION + it.timeToSlaAsString
		}.join("\n")
	}

	private def getSlaStatusText(def sla) {
		if(!sla.isStarted()) {
			i18n.getText("tts.common.slaCondition.inactive")
		} else if(sla.timeToSla > 0) {
			if(sla.finished) {
				i18n.getText("tts.common.slaConditions.met")
			} else if(sla.paused) {
				i18n.getText("tts.common.slaCondition.paused")
			} else if(sla.nonWorkingDay) {
				i18n.getText("tts.common.slaCondition.inNonWorkingDays")
			} else if(!sla.inWorkingTime) {
				i18n.getText("tts.common.slaCondition.inNonWorkingHours")
			} else {
				i18n.getText("tts.common.slaCondition.active")
			}
		} else {
			i18n.getText("tts.common.slaConditions.exceeded")
		}
	}

	// - Xray integration -----------------------------------------------------

	private def xrayManualTestStepsToString(def manualTestStepsFieldValue) {
		def testSteps = manualTestStepsFieldValue.steps
		testSteps.collect { step ->
			def actionLabel, action, data, result
			if (xray4OrLater) {
				actionLabel = i18n.getText('view-issue-test-section.field-manual-steps.step.label')
				action = step.fields.get("Action")
				data = "\n${i18n.getText('view-issue-test-section.field-manual-steps.data.label')}: ${step.fields.get('Data')}"
				result = "\n${i18n.getText('view-issue-test-section.field-manual-steps.expected.result.label')}: ${step.fields.get('Expected Result')}"
			} else {
				actionLabel = "Step"
				action = step.step
				data = step.data ? "\nData: ${step.data}" : ""
				result = step.result ? "\nExpected Result: ${step.result}" : ""
			}
			"${actionLabel} ${step.index}: ${action}${data}${result}"
		}.join("\n\n")
	}

	/**
	 * Returns the issue as "[FOO-123] My issue summary" or the passed issue key as "FOO-123",
	 * depending on the {@link #xrayIssuesWithSummaries} configuration variable.
	 */
	private def xrayIssueKeyValueToString(def issueKey) {
		xrayIssuesWithSummaries ? '[' + issueKey + '] ' + (issueManager.getIssueObject(issueKey)?.summary ?: '') : issueKey
	}
}
