import org.apache.commons.lang.StringUtils
import org.apache.log4j.Logger
import org.apache.lucene.search.Collector

import com.atlassian.jira.component.ComponentAccessor
import com.atlassian.jira.issue.search.SearchProvider
import com.atlassian.jira.jql.builder.JqlQueryBuilder
import com.atlassian.jira.user.ApplicationUser
import com.atlassian.query.Query

filename = "${title} - projects.xlsx".toString() // "filename" is a standard Java String, not a GString

projectExport = new ProjectExportTool()

/**
 * Velocity tool to export Jira project data.
 */
class ProjectExportTool {
	def log = Logger.getLogger(this.getClass())

	def archivedProjectManager
	def projectManager = ComponentAccessor.projectManager

	def projectIdToIssueCount = [:]
	def projectIdToUpdatedDate = [:]
	def projectTypeKeyToProjectTypeName = [:]

	def getAllProjects() {
		initArchivedProjectManager()
		initProjectStats()

		def allProjects = new ArrayList(projectManager.projectObjects)
		if (projectManager.metaClass.respondsTo(projectManager, "getArchivedProjects")) { // since Jira 7.10.0
			allProjects += projectManager.archivedProjects
		}

		allProjects.sort { it.key }
		return allProjects
	}

	def getStatus(project) {
		!project.hasProperty("archived") || !project.archived ? "Active" : "Archived" // since Jira 7.10.0
	}

	def getProjectType(project) {
		if (!project.hasProperty("projectTypeKey") || project.projectTypeKey == null) {
			return null
		}
		def projectTypeKey = project.projectTypeKey.key
		if (!projectTypeKeyToProjectTypeName.containsKey(projectTypeKey)) {
			projectTypeKeyToProjectTypeName[projectTypeKey] = formatProjectTypeKey(projectTypeKey)
		}
		return projectTypeKeyToProjectTypeName[projectTypeKey]
	}

	def getProjectCategory(project) {
		project.projectCategory?.name // since Jira 7.2.0
	}

	def getLastIssueUpdate(project) {
		unixTimeToExcelTime(projectIdToUpdatedDate[project.id])
	}

	def getIssueCount(project) {
		projectIdToIssueCount[project.id]
	}

	def getComponents(project) {
		project.components.collect { it.name }.join(", ")
	}

	def getIssueTypes(project) {
		project.issueTypes.collect { it.name }.join(", ")
	}

	def getVersions(project) {
		project.versions.collect { it.name }.join(", ")
	}

	def getArchivedDate(project) {
		unixTimeToExcelTime(archivedProjectManager?.getArchivedTimestamp(project))
	}

	def getArchivedBy(project) {
		archivedProjectManager?.getArchivedBy(project).value?.name
	}

	private def initArchivedProjectManager() {
		try {
			// since Jira 7.10.0
			archivedProjectManager = ComponentAccessor.getComponent(ComponentAccessor.classLoader.findClass("com.atlassian.jira.project.archiving.ArchivedProjectManager"))
		} catch (Exception ex) {
			log.warn("Failed to read ArchivedProjectManager", ex)
		}
	}

	private def initProjectStats() {
		try {
			def projectStats = getStatsForArchiving()
			projectIdToUpdatedDate = projectStats.mapToUpdatedDate
			projectIdToIssueCount = projectStats.mapToIssueCount
		} catch (Exception ex) {
			log.warn("Failed to read projects' statistics", ex)
		}
	}

	/**
	 * This is a modified version of the original method in <i>com.atlassian.jira.web.action.browser.BrowseProjects</i>.
	 */
	private def getStatsForArchiving() {
		def searchProvider = ComponentAccessor.getComponent(SearchProvider.class)
		def query = JqlQueryBuilder.newBuilder().buildQuery()

		def collector = ComponentAccessor.classLoader.findClass("com.atlassian.jira.web.action.browser.ProjectDataForArchivingCollector").newInstance()

		if (searchProvider.metaClass.respondsTo(searchProvider, "searchOverrideSecurity", Query, ApplicationUser, Collector)) {
			// since Jira 7.11.0
			searchProvider.searchOverrideSecurity(query, null, collector)
		} else {
			// since Jira 8
			def searchQueryClass = ComponentAccessor.classLoader.findClass("com.atlassian.jira.issue.search.SearchQuery")
			def searchQuery = searchQueryClass
					.getMethod("create", Query.class, ApplicationUser.class).invoke(null, query, null)
					.overrideSecurity(true)

			searchProvider.search(searchQuery, collector)
		}

		return collector
	}

	private def unixTimeToExcelTime(millis) {
		(millis != null) ? (25569 + millis / (24 * 60 * 60 * 1000)) : null
	}

	/**
	 * Same as in <i>com.atlassian.jira.project.type.ProjectTypeKeyFormatter</i>.
	 */
	private def formatProjectTypeKey(projectTypeKey) {
		projectTypeKey ? StringUtils.capitalize(StringUtils.replace(projectTypeKey, "_", " ").trim()) : "Unknown"
	}
}
