import org.apache.commons.io.IOUtils
import org.apache.log4j.Logger

def tempoApiToken = 'e4f41d78-f8e1-407a-9fba-ccafd2c83d86'

def tempoDateFromRequestParameter = requestContext?.getRequestParameter('tempoDateFrom')
def tempoDateToRequestParameter = requestContext?.getRequestParameter('tempoDateTo')
def tempoUsernameRequestParameter = requestContext?.getRequestParameter('tempoUsername')

def exportWorklogsOfAllTimes = !(tempoDateFromRequestParameter && tempoDateToRequestParameter)
def exportWorklogsOfAllUsers = false // whether to export the worklog of all users or only the one specified by tempoUsernameRequestParameter

def tempoDateFrom = tempoDateFromRequestParameter ?: '2000-01-01'
def tempoDateTo = tempoDateToRequestParameter ?: '2099-12-31'
def tempoUsername = !exportWorklogsOfAllUsers ? tempoUsernameRequestParameter : null

tempo = new TempoTool(apiToken: tempoApiToken, baseUrl: requestContext.canonicalBaseUrl, dateFrom: tempoDateFrom, dateTo: tempoDateTo, username: tempoUsername, exportWorklogsOfAllTimes: exportWorklogsOfAllTimes, issues: issues)

/**
 * Velocity tool to query Tempo worklogs from the Tempo Servlet API.
 *
 * @see https://tempoplugin.jira.com/wiki/display/TEMPO/Tempo+Servlet+Manual
 */
class TempoTool {
	def log = Logger.getLogger(this.getClass())

	/* Tempo servlet request parameters. */
	def apiToken
	def baseUrl

	def dateFrom
	def dateTo
	/**
	 * If not <code>null</code>, only the worklogs of this user will be returned.
	 */
	def username

	def exportWorklogsOfAllTimes
	def issues

	/**
	 * Whether to preload all the worklogs in given timeframe with a single Tempo API call.
	 * If false, worklogs will be downloaded for each issue separately.
	 * The former performs better for exports with larger number of issues.
	 */
	def preloadWorklogs
	def preloadedWorklogs = []

	def slurper = new XmlSlurper()

	/**
	 * Returns the Tempo managed worklogs for the passed issue as a <code>List</code> of <code>Maps</code>, or the exception object if something went wrong.
	 * You can trivially iterate over the list and access the values by the keys visible in the "def m = ..." line.
	 * Ex: "hours", "work_date", etc.
	 */
	def getWorklogs(def issueKey) {
		preloadWorklogs = (issues.size() > 10) && !exportWorklogsOfAllTimes
		if(preloadWorklogs) {
			if(!preloadedWorklogs) {
				log.debug("Downloading Tempo worklogs in bulk")
				preloadedWorklogs = callTempoApi()
			}
			return preloadedWorklogs.findAll { it.issue_key == issueKey }
		} else {
			log.debug("Downloading Tempo worklogs for ${issueKey}")
			return callTempoApi(issueKey)
		}
	}

	/**
	 * Calls the Tempo API and returns a list of objects with predefined worklog attributes
	 * @param issueKey optional, if not defined, then all worklogs are returned in set timeframe
	 */
	def callTempoApi(def issueKey) {
		InputStream stream
		try {
			// note: you can't filter by username (the Tempo API can filter either by issue key or by username, but not both)
			def url = "${baseUrl}/plugins/servlet/tempo-getWorklog/?dateFrom=${dateFrom}&dateTo=${dateTo}&format=xml&addBillingInfo=true&addUserDetails=true&tempoApiToken=${apiToken}"
			if(issueKey) {
				url += "&issueKey=${issueKey}"
			} else if(username) {
				url += "&userName=${username}"
			}
			log.debug("Tempo URL" + (issueKey ? " for ${issueKey}" : "") + ": ${url}")
			stream = new URL(url).openStream()
			def xml = IOUtils.toString(stream, "UTF-8")
			log.debug("Tempo XML" + (issueKey ? " for ${issueKey}" : "") + ":\n${xml}")

			def result = slurper.parseText(xml)
			def worklogs = result.worklog
			log.debug("${worklogs.size()} Tempo worklogs found" + (issueKey ? " for ${issueKey}" : ""))

			// deserialize
			def worklogObjects = []
			result.worklog.each { w ->
				if((issueKey && username) && (username != w.username.text())) { // w.username is an instance of groovy.util.slurpersupport.NodeChildren
					return
				}
				def m = [ issue_key: w.issue_key, hours: w.hours.toDouble(), work_date: Date.parse("yyyy-MM-dd", w.work_date as String), username: w.username, billing_key: w.billing_key, work_description: w.work_description, full_name: w.user_details.full_name, billing_name: w.Billing.@name, activity_name: w.activity_name ]
				worklogObjects.add(m)
			}
			log.debug("Worklogs" + (issueKey ? " for ${issueKey}" : "") + ": ${worklogObjects}")
			return worklogObjects
		} catch(Throwable ex) {
			log.error('Failed to download Tempo worklogs', ex)
			def failureMessage = "Failed to download Tempo worklogs (have you enabled the Tempo API?): ${ex.message}"
			return [
				[ issue_key: "", hours: "", work_date: "", username: "", billing_key: "", work_description: failureMessage, full_name: "", billing_name: "", activity_name: "" ]
			]
		} finally {
			IOUtils.closeQuietly(stream)
		}
	}
}
